// A volume (= 3D matrix, 3D image) class, with a few tools
// Simple loading-from-file and display (text mode)
// For more practical loading/saving (from/to png, jpg, ... files) or display (color pixels in a window), see with_graphical_library.h, which is based on an external graphical library, and... adapt it :)

#ifndef VOLUMES_H
#define VOLUMES_H

#include <cmath>
#include <unistd.h>
#include <algorithm>
#include <cassert>

#include <fstream>
#include <string>
#include <iostream>
using namespace std;


namespace MRFSegm {

// A class Volume, to contain a 3D-table of int, double or anything else.

template <class precis>
struct Volume {

  int nx, ny, nz;
  int nxy; // to avoid recomputation of nx*ny
  precis * val;

  // constructor

  Volume(){
    nx = 0; ny = 0; nz = 0; nxy = 0;
  }

  // constructor allocating the size required

  Volume(int x, int y, int z){
    allocate(x,y,z);
  }

  // constructor copying a pointer to the table
  Volume(precis* tab, int x, int y, int z){
    nx = x; ny = y; nz = z; nxy = nx * ny; val = tab;
  }

  // constructor initializing with an image from a file, and applying a zoom factor

  Volume(const char * fich, double zoom = 1);
  void load(const char* fich, double zoom = 1);
  void load(const string fich, double zoom = 1){
    load(fich.c_str(), zoom);
  }

  // manual desctructor ... not automatic !

  void empty(){
    if ((nx!=0)&&(ny!=0)&&(nz!=0)){
      nx = 0; ny = 0; nz = 0; nxy = 0;
      delete[] val;
    }
  }  

  /*
  // why not ? because of pointers shared for certain types of copies.
  ~Volume(){ empty(); }
  */

  // To allocate by hand

  void allocate(int x, int y, int z){
    nx = x; ny = y; nz = z; nxy = nx * ny;
    if ((x!=0)&&(y!=0)&&(z!=0)){
      try {
	val = new precis[nx*ny*nz]; 
      } 
      catch (bad_alloc& ba) {
	cerr << "Bad alloc in volume allocation, size " << x <<"x"<< y <<"x"<< z << " : " << ba.what() << endl;
      }
    }
  }

  // To handle the table easily and carfully
  
  inline precis& operator()(int x, int y, int z){
    assert((x<nx)&&(y<ny)&&(z<nz)&&(x>=0)&&(y>=0)&&(z>=0));
    return val[x + nx*y + nxy*z];
  }

  inline precis operator()(int x, int y, int z) const {
    assert((x<nx)&&(y<ny)&&(x>=0)&&(y>=0));
    return val[x + nx*y + nxy*z];
  }

  void operator+=(const Volume<precis>& B){
    assert(nx == B.nx); assert(ny == B.ny); assert(nz == B.nz); 
    for (int i = 0; i<nx; i++)
      for (int j = 0; j<ny; j++)
	for (int k = 0; k<nz; k++)
	  (*this)(i,j,k) += B(i,j,k);
  }

  void operator-=(const Volume<precis>& B){
    assert(nx == B.nx); assert(ny == B.ny); assert(nz == B.nz); 
    for (int i = 0; i<nx; i++)
      for (int j = 0; j<ny; j++)
	for (int k = 0; k<nz; k++)
	  (*this)(i,j,k) -= B(i,j,k);
  }

  Volume<precis> operator+(const Volume<precis>& B) const {
    assert(nx == B.nx); assert(ny == B.ny); assert(nz == B.nz); 
    Volume<precis> C(nx,ny,nz);
    for (int i = 0; i<nx; i++)
      for (int j = 0; j<ny; j++)
	for (int k = 0; k<nz; k++)
	  C(i,j,k) = (*this)(i,j,k) + B(i,j,k);
    return C;
  }

  Volume<precis> operator-(const Volume<precis>& B) const {
    assert(nx == B.nx); assert(ny == B.ny); assert(nz == B.nz); 
    Volume<precis> C(nx,ny,nz);
    for (int i = 0; i<nx; i++)
      for (int j = 0; j<ny; j++)
	for (int k = 0; k<nz; k++)
	  C(i,j,k) = (*this)(i,j,k) - B(i,j,k);
    return C;
  }

  // To have test images
  void fill(precis p); // fill with value p
  void nul();  // all coefficients to value 0

  // To read an image from a simple text format file (see code)
  void read_image_from_textfile(const char * fich);

  // Display with possibly a zoom factor and a translation

  void display(double zoom = 1, int p = 0, int q = 0);
  void save(const char * fich, double zoom = 1);
  void save(string fich, double zoom = 1){
    save(fich.c_str(), zoom);
  }
  void display_text(int prec = -1) const;

};


// Common types

typedef Volume<int> IntVolume;
typedef Volume<double> DoubleVolume;
//typedef Volume<float> FloatVolume;
//typedef Volume<float> DoubleVolume;
typedef Volume<bool> BoolVolume;


// A few declarations

template <class precis1, class precis2>
void copy(const Volume<precis1>& A, Volume<precis2>& B);


// already included in images.h
// Basic maths
//inline int square(int a){ return (a*a);}
//inline double square(double a){ return (a*a);}

// Useful tool
//inline int limite(int a, int b, int c){
//  if (b<a) return a;
//  if (b>c) return c;
//  return b;
//}

//inline int limite(int a, double b, int c){
//  return limite(a,int(b),c);
//}



// Read an image from a file in an homemade text format
// Put the full path + name of file in 'fich'
// To read a png, jpg (etc.) image from a file, you should use with_graphical_library.h, which makes use of an external graphical library
template <class precis> 
void Volume<precis>::read_image_from_textfile(const char * fich) {
  //  string fichier = "/users/odyssee/charpiat/recalage/images/" + string(fich) + ".dat";
  string fichier(fich);
  ifstream f;
  f.open(fichier.c_str());
  f >> nx;
  f >> ny;
  f >> nz;
  cout << "Size of the image loaded: " << nx << " x " << ny << " x " << nz << endl;
  allocate(nx, ny, nz);
  for (int i=0; i<nx; i++)
    for (int j = 0; j<ny; j++)
      for (int k = 0; k<nz; k++){
	f >> (*this)(i,j,k);
	//      cout << U(i,j);
      }

}



// Displays the matrix in text mode in the terminal
// prec = nb of digits
// To display images in a window (as pixels), you should use with_graphical_library.h, which makes use of an external graphical library
template <class T>
void Volume<T>::display_text(int prec) const{
  cout << endl << "Size: " << nx << " x " << ny <<  " x " << nz <<endl;

  for (int k = 0; k<nz; k++){ // displaying nz images successively

    for (int i = 0; i<nx; i++){
      int nbd =  int( ceil( log(nx)/log(10) ) );
      char* b = new char[nbd+1];
      sprintf(b, "%*d", nbd, i);
      cout << "L " << string(b) << " :  ";
      delete[] b;
      
      for (int j = 0; j<ny; j++){
	if (prec == -1){ // standard case
	  cout << "  " << (*this)(i,j,k);
      } else { // case with alignment
	  char* c = new char[prec+5];
	  sprintf(c, "% -*.*g", prec+2, prec, double ( (*this)(i,j,k) ) );
	  cout << "  	" << string(c);
	  delete[] c;
	}
      }
      cout << endl;
    }

    cout << endl;
  }
}






// Fill with 0
template <class precis>
void Volume<precis>::nul(){
  for (int i = 0; i < nx; i++)
    for (int j = 0; j < ny; j++)
      for (int k = 0; k < nz; k++)
	(*this)(i,j,k) = precis(0);
}


// Multiply an image by a constant factor

template <class precis1>
void multiply(Volume<precis1>& A, double z){
  for (int i = 0; i<A.nx; i++)
    for (int j = 0; j<A.ny; j++)
      for (int k = 0; k<A.nz; k++)
	A(i,j,k) *= z;
}


// Copies element by element a matrix into a new one.
template <class precis1, class precis2>
void copy(const Volume<precis1>& A, Volume<precis2>& B){
  if ((B.nx != A.nx)||(B.ny != A.ny)||(B.nz != A.nz)){
    B.empty();
    B.allocate(A.nx, A.ny, A.nz);
  }

  for (int i = 0; i<A.nx; i++)
    for (int j = 0; j<A.ny; j++)
      for (int k = 0; k<A.nz; k++)
	B(i,j,k) = precis2(A(i,j,k));
}

// Copies and deletes the original one
template <class precis>
void copy_suppr(Volume<precis>& A, Volume<precis>& B){
  B.empty();
  B.val = A.val;
  B.nx = A.nx;
  B.ny = A.ny;
  B.nz = A.nz;
  B.nxy = A.nxy;
  A.nx = 0;
  A.ny = 0;
  A.nz = 0;
  A.nxy = 0;
  A.val = NULL;
}


// Max norm (infinity norm) of a matrix
double norm_max(const DoubleVolume& A){
  double z = 0;
  double nor;
  for (int i =0; i<A.nx; i++)
    for (int j =0; j<A.ny; j++)
      for (int k =0; k<A.nz; k++){
	nor = fabs( A(i,j,k) );
	if (nor>z) z = nor;
      }
  return z;
}

// Smallest coefficient in the matrix
double smallest_coeff(const DoubleVolume& A){
  double z = A(0,0,0);

  for (int i = 0; i<A.nx; i++)
    for (int j = 0; j<A.ny; j++)
      for (int k = 0; k<A.nz; k++)
	if (A(i,j,k) < z) z = A(i,j,k);

  return z;
}



// max of three numbers...
template <class T>
inline T max3(const T& a, const T& b, const T& c){
  return max(a, max(b, c));
}


// Smooth an image/matrix with a Gaussian blur 
// Case of matrix of integers
void smooth(IntVolume& A, double sigma){
  DoubleVolume B;
  copy(A, B);
  
  double gamma = 1/(2*sigma*sigma);
  int larg = min(2*max3(A.nx, A.ny, A.nz), 2*max(1,int(sigma))+1);

  int zero = larg/2;
  int tx = A.nx; int ty = A.ny; int tz = A.nz;
  
  DoubleVolume masque(larg, larg, larg);
  
  for (int i =0; i<larg; i++)
    for (int j = 0; j < larg; j++)
      for (int k = 0; k < larg; k++)
	masque(i,j,k) = exp(-gamma*sqrt(double(square(i-zero)+square(j-zero)+square(k-zero))));
  
  for (int i=0; i<tx; i++)
    for (int j =0; j<ty; j++)
      for (int k =0; k<tz; k++){
      
      int pmin = limite(0,i-zero,tx-1);
      int pmax = limite(0,i+zero,tx-1);
      int qmin = limite(0,j-zero,ty-1);
      int qmax = limite(0,j+zero,ty-1);
      int rmin = limite(0,k-zero,tz-1);
      int rmax = limite(0,k+zero,tz-1);
      
      B(i,j,k) = 0;
      double poids = 0;
      for (int p = pmin; p<=pmax; p++)
	for(int q = qmin; q<=qmax; q++)
	  for(int r = rmin; r<=rmax; r++){
	    double poids_aux = masque(p-i+zero, q-j+zero, r-k+zero);
	    poids += poids_aux;
	    B(i,j,k) += poids_aux * A(p,q,r);
	} 
	  B(i,j,k) /= poids;
    }  
  
  copy(B, A);
  B.empty();
  masque.empty();
}

// Case of matrix of doubles
void smooth(DoubleVolume& A, double sigma, double coeffsigma = 3){

  DoubleVolume B;
  copy(A, B);
  
  double gamma = 1/(2*sigma*sigma);
  int larg = min(2*max3(A.nx, A.ny, A.nz), int(2*coeffsigma*max(1,int(sigma))+1));

  int zero = larg/2;
  int tx = A.nx; int ty = A.ny; int tz = A.nz;
  
  DoubleVolume masque(larg, larg, larg);
  
  for (int i =0; i<larg; i++)
    for (int j = 0; j < larg; j++)
      for (int k = 0; k < larg; k++)
	masque(i,j,k) = exp(-gamma*sqrt(double(square(i-zero)+square(j-zero)+square(k-zero))));
  
  for (int i=0; i<tx; i++)
    for (int j =0; j<ty; j++)
      for (int k =0; k<tz; k++){
      
      int pmin = limite(0,i-zero,tx-1);
      int pmax = limite(0,i+zero,tx-1);
      int qmin = limite(0,j-zero,ty-1);
      int qmax = limite(0,j+zero,ty-1);
      int rmin = limite(0,k-zero,tz-1);
      int rmax = limite(0,k+zero,tz-1);
      
      B(i,j,k) = 0;
      double poids = 0;
      for (int p = pmin; p<=pmax; p++)
	for(int q = qmin; q<=qmax; q++)
	  for(int r = rmin; r<=rmax; r++){
	    double poids_aux = masque(p-i+zero, q-j+zero, r-k+zero);
	    poids += poids_aux;
	    B(i,j,k) += poids_aux * A(p,q,r);
	} 
	  B(i,j,k) /= poids;
    }  
  
  copy(B, A);
  B.empty();
  masque.empty();
}


// Copies with a zoom factor (a bit rough)
template <class precis, class precis2>
void copy_fact(Volume<precis>& A, Volume<precis2>& B, double zoom){

  B.empty();
  int tx = int(A.nx * zoom); 
  int ty = int(A.ny * zoom);
  int tz = int(A.nz * zoom);
  B.allocate(tx, ty, tz);
  
  double zx = double(A.nx)/B.nx;
  double zy = double(A.ny)/B.ny;
  double zz = double(A.nz)/B.nz;

  for (int i = 0; i<B.nx; i++)
    for (int j = 0; j<B.ny; j++)
      for (int k = 0; k<B.nz; k++)
	B(i,j,k) = precis2(A(limite(0,zx*i,A.nx-1), limite(0,zy*j,A.ny-1), limite(0,zz*k,A.nz-1)));
}


// Fill with a constant value (supposes allocation has already been done)
template <class precis>
void Volume<precis>::fill(precis p){
  for (int i = 0; i<nx; i++)
    for (int j = 0; j<ny; j++)
      for (int k = 0; k<nz; k++)
	(*this)(i,j,k) = p;
}


// Infinite norm between 2 matrices
// T : int, double, float... or else add .norme()
template <class T>
double difference_max(const Volume<T>& A, const Volume<T>& B){
  double x = 0;
  assert(A.nx == B.nx);
  assert(A.ny == B.ny);
  assert(A.nz == B.nz);
  for (int i = 0; i<A.nx; i++)
    for (int j = 0; j<A.ny; j++)
      for (int k = 0; k<A.nz; k++){
	double dx = fabs(A(i,j,k) - B(i,j,k));
	if (dx > x)
	  x = dx;
      }
 return x;
}



// Returns a slice
// cut direction : 0=x, 1=y, 2=z;
// cut location

 template <class T>
 RecImage<T> slice(const Volume<T>& V, int dir, int icut){
   int da, db; // a,b : coordinates on the slice (x y or z)
   switch (dir){
   case 0:
     da = 1; db = 2;
     break;
   case 1:
     da = 0; db = 2;
     break;
   case 2:
     da = 0; db = 1;
     break;
   default:
     cout << "Come on, guys..." << endl;
   }

   int taille[3]; taille[0] = V.nx; taille[1] = V.ny; taille[2] = V.nz;
   int na = taille[da]; int nb = taille[db];

   RecImage<T> tranche(na, nb);

   int coord[3]; coord[0] = icut; coord[1] = icut; coord[2] = icut;

   for (int a = 0; a<na; a++)
     for (int b = 0; b<nb; b++){
       coord[da] = a; coord[db] = b;
       tranche(a,b) = V(coord[0], coord[1], coord[2]);
     }
   
   return tranche;
 }
 

 

} // end of namespace
#endif
