// Structures, types and basic functions for :
// Segmente une suite d'image avec la contrainte que la forme dans l'image t contient la forme de l'image t+1.

#ifndef VIDEO_CUT_AVEC_CONTRAINTE_INCLUSION
#define VIDEO_CUT_AVEC_CONTRAINTE_INCLUSION

// A few standard includes
#include <string>
#include <sstream>
#include <fstream>
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
using namespace std;

#include <cmath>
#include <list>
#include <vector>
#include <algorithm>
#include <cassert>

// Personal libraries
#include "images.h"
#include "volumes.h"
#include "randomness.h"


// MRF - graph-cut by Kolmogorov and co
//#include "mrf.h"
//#include "ICM.h"
//#include "GCoptimization.h"
//#include "MaxProdBP.h"
//#include "graph.h"

namespace MRFSegm {

/************************************************************************/
/************************  Boring tools  ********************************/
/************************************************************************/


/******************** Display *******************/

// Displays a vector in a terminal
template <class T>
void display_text(const vector<T>& V){

  typename vector<T>::const_iterator ita = V.begin(); 

  for ( ; ita != V.end(); ++ita)
    cout << (*ita) << "  ";
  cout << endl;
}


// to print a pair
template <class A, class B>
ostream& operator<<(ostream& F, const pair<A,B>& P){
  return F << "(" << P.first << ", " << P.second << ")";
}


// to print a list
template <class A>
ostream& operator<<(ostream& F, const list<A>& P){
  for (typename list<A>::const_iterator i = P.begin(); i != P.end(); ++i)
    F << (*i) << ", " ;
  return F;
}

// to print a vector<>
template <class A>
ostream& operator<<(ostream& F, const vector<A>& P){
  F << "[";
  for (typename vector<A>::const_iterator i = P.begin(); i != P.end(); ++i)
    F << (*i) << "; " ;
  F << "]";
  return F;
}

// to print a DoubleImage
template <class A>
ostream& operator<<(ostream& F, const RecImage<A>& M){
  int prec = 5;
  F << endl << "Size: " << M.nx << " x " << M.ny << endl;
  for (int i = 0; i<M.nx; i++){
    int nbd =  int( ceil( log(M.nx)/log(10) ) );
    char* b = new char[nbd+1];
    sprintf(b, "%*d", nbd, i);
    F << "L " << string(b) << " :  ";
    delete[] b;
    
    for (int j = 0; j<M.ny; j++){
      if (prec == -1){ // standard case
        F << "  " << M(i,j);
      } else { // case with alignment
        char* c = new char[prec+5];
        sprintf(c, "% -*.*g", prec+2, prec, double ( M(i,j) ) );
        F << "       " << string(c);
        delete[] c;
      }
    }
    F << endl;
  }
  return F;
}


// to print a DoubleVolume
template <class A>
ostream& operator<<(ostream& F, const Volume<A>& M){
  int prec = 5;
  F << endl << "Size: " << M.nx << " x " << M.ny << " x " << M.nz << endl;

  // print each z-slice, one by one
  for (int k = 0; k<M.nz; k++){

    F << "Slice z = " << k << " : " << endl;

    for (int i = 0; i<M.nx; i++){
      int nbd =  int( ceil( log(M.nx)/log(10) ) );
      char* b = new char[nbd+1];
      sprintf(b, "%*d", nbd, i);
      F << "L " << string(b) << " :  ";
      delete[] b;
      
      for (int j = 0; j<M.ny; j++){
	if (prec == -1){ // standard case
	  F << "  " << M(i,j,k);
	} else { // case with alignment
	  char* c = new char[prec+5];
	  sprintf(c, "% -*.*g", prec+2, prec, double ( M(i,j,k) ) );
	  F << "       " << string(c);
	  delete[] c;
	}
      }
      F << endl;
    }
  }
  F << endl;

  return F;
}



/**************** Simple functions ****************/

// Swaps 2 elements of a table 
template <class T>
void swap_elements(vector<T>& V, int i, int j){
  T a = V[i];
  V[i] = V[j];
  V[j] = a;
}




/******************************* Random stuff ***************************/

// Generates a random matrix with coefficients between 0 and 1 (uniform law)
DoubleImage random_matrix(int n, int m, double coeff = 1.){
  DoubleImage M(n,m);
  for (int i = 0; i<n; i++)
    for (int j = 0; j<m; j++)
      M(i,j) = rand() * coeff / RAND_MAX;
      //      M(i,j) = (rand() % 2 )* 1. / 2;

  return M;
}


// Generates a random vector with coefficients between 0 and 1 (uniform law)
vector<double> random_vector(int n, double coefficient = 1.){
  vector<double> V(n);
  for (vector<double>::iterator i = V.begin(); i != V.end(); ++i)
    (*i) = rand() * coefficient / RAND_MAX;

  return V;
}





/************************************************************************************/
/********************************** MRF representation ******************************/
/************************************************************************************/


/**********************  ImageEnergy  **************************/

// Structure that encodes a binary MRF on one image
// T : int, double, float...
template <class T>
struct ImageEnergy {

  // data structure
  int nx; // width
  int ny; // height 

  int neighborhoodsize; // in the actual version of the code, has to be 4 or 8 (greatly recommanded : 8)

  // energy description
  RecImage< T > Pot; // the potential (of being the iceberg = the foreground object) for each variable. Can be any number, positive, zero, negative... Pot(x,y) is the cost, for point (x,y), to be part of the iceberg in this image  -  the cost of being part of the background.

  RecImage< vector<T> > Inter; // the interaction term between neighborhing pixels. Inter[x][d] is the link force between pixel x in direction d (see below)

  //  .    Voisinage à 4 points
  // .x.
  //  . 
  //
  // ...   Voisinage à 8 points (fortement recommandé)
  // .x.
  // ...

  // 
  //  5 6 7   le système de voisinage à 8 voisins autour du point x 
  //  4 x 0   On stocke la force du lien avec les voisins 0 1 2 3 UNIQUEMENT, 
  //  3 2 1   dans Inter[x][0], Inter[x][1]. Inter[x][2] et Inter[x][3],
  //          Les liens avec les autres voisins sont déjà décrits par ceux-ci (symétrie) :
  //
  //                                    ...
  // y.. Le lien xy est déjà décrit par .y. dans Inter[y][1]
  // .x.     Inter[x][5] ? Non !        ..x
  // ...


  // energy extras
  double constante; // additional constant to the energy


  // constructors
  ImageEnergy(){
    nx = 0; ny = 0;
    neighborhoodsize = 8;
    constante = 0;
  }

  void allocate(int x, int y){ // don't forget to empty first if not empty...
    nx = x; ny = y;
    Pot.allocate(nx, ny);
    Inter.allocate(nx,ny);
    for (int i = 0; i<nx; i++)
      for (int j = 0; j<ny; j++)
	Inter(i,j).resize(neighborhoodsize/2);
    constante = 0;
  }

  ImageEnergy(int x, int y){
    neighborhoodsize = 8;
    allocate(x,y);
  }

  ImageEnergy(int x, int y, int neigh){
    neighborhoodsize = neigh;
    allocate(x,y);
  }

  ImageEnergy(RecImage<T>& P, RecImage< vector<T> >& I) : Pot(P), Inter(I) { // attention, copie de pointeurs, données partagées !
    neighborhoodsize = 8;
    nx = Pot.nx; ny = Pot.ny;
    assert(Inter.nx == nx);
    assert(Inter.ny == ny);
    constante = 0;
    if ((nx != 0) && (ny != 0)) 
      neighborhoodsize = 2 * Inter[0].size;
  }

  void initialise_energy(const RecImage<T>& P, const RecImage< vector<T> >& I){ // copie des données (temps perdu à copier)
    assert(P.nx == I.nx);
    assert(P.ny == I.ny);
    if ( (P.nx != nx) || (P.ny != ny) )
      allocate(P.nx,P.ny); // actually not mandatory (otherwisely automatically done by next line)
    copy(P, Pot); 
    copy(I, Inter);
  }

  void initialise_energy_suppr(RecImage<T>& P, RecImage< vector<T> >& I){ // copie les pointeurs (supprimer l'original, pas de temps perdu)
    assert(P.nx == I.nx);
    assert(P.ny == I.ny);
    copy_suppr(P, Pot); 
    copy_suppr(I, Inter);
  }

  void empty(){
    Pot.empty();
    Inter.empty();
    nx = 0; ny = 0;
  }    

  ~ImageEnergy(){ // ne fait rien. Il vaut mieux vider à la main pour éviter de supprimer des pointeurs dans le cas d'initialisation par copie de pointeur
    //    empty();
  }

  // Miscellaneous
  void display_text() const{
    cout << "Image energy : " << nx << " x " << ny << " pixels." << endl;
    cout << "Constant : " << constante << endl;
    cout << "Pot :  "; MRFSegm::display_text(Pot);
  }

};



/**********************  VideoEnergy  **************************/


// Structure that encodes a binary MRF on a sequence of (independant) images
// T : int, double, float...
template <class T>
struct VideoEnergy {

  // data structure
  int nt; // number of frames

  // data structure information already given in each Pot[t]
  int nx; // width
  int ny; // and height of each image (all images have the same size)

  int neighborhoodsize; // in the actual version of the code, has to be 4 or 8 (greatly recommanded : 8)


  // energy description
  vector< ImageEnergy<T> > Frames; // the energies defined independently on each frame. For instance Frames[t].Pot(x,y) is the cost, for point (x,y), to be part of the iceberg in image t  -  the cost of being part of the background.

  // energy extras
  double constante; // additional constant to the total energy, which is the sum of the energies of each frame


  // constructors 
  VideoEnergy(){
    nt = 0; 
    nx = 0; ny = 0;
    constante = 0;
    neighborhoodsize = 8;
  }

  void allocate(int nnt){
    Frames.resize(nnt);
    nt = nnt;
  }

  VideoEnergy(int nnt){
    allocate(nnt);
    nx = 0; ny = 0;
    constante = 0;
    neighborhoodsize = 8;
  }

  VideoEnergy(int nnt, int nnx, int nny, int neigh = 8){
    allocate(nnt);
    nx = nnx; ny = nny;
    constante = 0;
    neighborhoodsize = neigh;
    
    for (int t = 0; t<nt; t++){
      Frames[t].neighborhoodsize = neighborhoodsize;
      Frames[t].allocate(nx, ny);
    }
  }

  VideoEnergy(const vector< ImageEnergy<T> >& FFrames) : Frames(FFrames){ // attention, copie des pointeurs sur les données
    nt = Frames.size();
    nx = 0; ny = 0; neighborhoodsize = 8; if (nt !=0){ nx = Frames[0].nx; ny = Frames[0].ny; neighborhoodsize = Frames[0].neighborhoodsize;}
    constante = 0;
  }


  ~VideoEnergy(){} // destructor by default 

  void empty_light(){ // doesn't clean much (useful if initialised by copying pointers to data which should not be deleted)
    Frames.empty();
  }

  void empty_full(){ // if really needed
    for (int t = 0; t<nt; t++)
      Frames[t].empty();
    Frames.empty();
  }


  void display_text() const{
    
    cout << "Video energy : " << nt << " frames of " << nx << " x " << ny << " pixels." << endl;
    cout << "Additional constant : " << constante << endl << endl;
  }

  
};

 
 
 
// Evaluates the energy associated to a state (a choice of all labels)
template <class T>
double evaluate_energy(const ImageEnergy<T>& A, const RecImage<bool>& S){
  double v = A.constante;
  int ns = A.neighborhoodsize/2;
  assert(S.nx == A.nx);
  assert(S.ny == A.ny);
  for (int i = 0; i<A.nx; i++)
    for (int j = 0; j<A.ny; j++){
      if (! S(i,j))
	v += A.Pot(i,j); 
  //      else v += 0

      for (int k = 0; k<ns; k++){
	int i0 = i; int j0 = j;

	switch (k) {

	  case 0 :
	    if (i == A.nx-1)
	      continue;
	    i0 += 1;
	    break;

	  case 1 :
	    if ( (i == A.nx-1) || (j == A.ny - 1))
	      continue;
	    i0 += 1;
	    j0 += 1;
	    break;

	  case 2 :
	    if (j == A.ny - 1)
	      continue;
	    j0 += 1;
	    break;

	  case 3 :
	    if ( (i == 0) || (j == A.ny - 1))
	      continue;
	    i0 -= 1;
	    j0 += 1;
	    break;

	  default:
	    cout << "Vous êtes envahis par vos voisins!" << endl;
	  }

	if (S(i,j) != S(i0,j0))
	  v += A.Inter(i,j)[k];
      }
      
    }
  
  return v;
 }
 

// Evaluates the energy associated to a state (a choice of all labels)
template <class T>
double evaluate_energy(const VideoEnergy<T>& A, const vector< RecImage<bool> >& S){
  double v = A.constante;
  for (int k = 0; k<A.nt; k++)
    v += evaluate_energy(A.Frames[k], S[k]);
  return v;
}








/************************************************************************************/
/********************************** The same but in 3D ******************************/
/************************************************************************************/


/**********************  Image3DEnergy  **************************/

// Structure that encodes a binary MRF on one image
// T : int, double, float...
template <class T>
struct VolumeEnergy {

  // data structure
  int nx; // width
  int ny; // height 
  int nz; // depth

  int neighborhoodsize; // in the actual version of the code, has to be 6 or 26 (greatly recommanded : 26)

  // energy description
  Volume< T > Pot; // the potential (of being the iceberg = the foreground object) for each variable. Can be any number, positive, zero, negative... Pot(x,y,z) is the cost, for point (x,y,z), to be part of the iceberg in this image  -  the cost of being part of the background.

  Volume< vector<T> > Inter; // the interaction term between neighborhing pixels. Inter[x][d] is the link force between pixel x in direction d (see below)

  //       .         Voisinage à 6 points (+- 1 sur x y ou z)
  //  .   .x.   .      (faces d'un cube)
  //       . 
  //
  //   ... ... ...  Voisinage à 26 points (fortement recommandé)(+- 1 sur x y et/ou z)
  //   ... .x. ...     (Rubik's cube)
  //   ... ... ...
  //dessous /  dessus

  //    
  //  0 1 2     9 10 11   18 19 20   Voisinage à 26 points
  //  3 4 5    12  x 14   21 22 23   (le numéro 13 = le pixel courant, n'est pas utilisé)
  //  6 7 8    15 16 17   24 25 26   
  //
  //   z-1        z         z+1
  //  direction d = (dx, dy, dz) avec chaque di = -1, 0 ou 1   --> indice_direction = (dx+1) + (dy+1)*3 + (dz+1)*9
  //
  // Dans .Inter, on ne stocke que les interactions <= 12
  // car chaque interaction entre deux pixels ne doit être représentée qu'une fois (soit par lui, soit par son voisin).
  // Exemple d'utilisation : Inter[x][12]



  // energy extras
  double constante; // additional constant to the energy


  // constructors
  VolumeEnergy(){
    nx = 0; ny = 0; nz = 0;
    neighborhoodsize = 26;
    constante = 0;
  }

  void allocate(int x, int y, int z){ // don't forget to empty first if not empty...
    nx = x; ny = y; nz = z;
    Pot.allocate(nx, ny, nz);
    Inter.allocate(nx,ny, nz);
    for (int i = 0; i<nx; i++)
      for (int j = 0; j<ny; j++)
	for (int k = 0; k<nz; k++)
	  Inter(i,j,k).resize(neighborhoodsize/2);
    constante = 0;
  }

  VolumeEnergy(int x, int y, int z){
    neighborhoodsize = 26;
    allocate(x,y,z);
  }

  VolumeEnergy(int x, int y, int z, int neigh){
    neighborhoodsize = neigh;
    allocate(x,y,z);
  }

  VolumeEnergy(Volume<T>& P, Volume< vector<T> >& I) : Pot(P), Inter(I) { // attention, copie de pointeurs, données partagées !
    neighborhoodsize = 26;
    nx = Pot.nx; ny = Pot.ny; nz = Pot.nz;
    assert(Inter.nx == nx);
    assert(Inter.ny == ny);
    assert(Inter.nz == nz);
    constante = 0;
    if ((nx != 0) && (ny != 0) && (nz != 0)) 
      neighborhoodsize = 2 * Inter[0].size;
  }

  void initialise_energy(const Volume<T>& P, const Volume< vector<T> >& I){ // copie des données (temps perdu à copier)
    assert(P.nx == I.nx);
    assert(P.ny == I.ny);
    assert(P.nz == I.nz);
    if ( (P.nx != nx) || (P.ny != ny) || (P.nz != nz))
      allocate(P.nx,P.ny,P.nz); // actually not mandatory (otherwisely automatically done by next line)
    copy(P, Pot); 
    copy(I, Inter);
  }

  void initialise_energy_suppr(Volume<T>& P, Volume< vector<T> >& I){ // copie les pointeurs (supprimer l'original, pas de temps perdu)
    assert(P.nx == I.nx);
    assert(P.ny == I.ny);
    assert(P.nz == I.nz);
    copy_suppr(P, Pot); 
    copy_suppr(I, Inter);
  }

  void empty(){
    Pot.empty();
    Inter.empty();
    nx = 0; ny = 0; nz = 0;
  }    

  ~VolumeEnergy(){ // ne fait rien. Il vaut mieux vider à la main pour éviter de supprimer des pointeurs dans le cas d'initialisation par copie de pointeur
    //    empty();
  }

  // Miscellaneous
  void display_text() const{
    cout << "Volume energy : " << nx << " x " << ny << " x " << nz << " pixels." << endl;
    cout << "Constant : " << constante << endl;
    cout << "Pot :  "; MRFSegm::display_text(Pot);
  }

};



/**********************  VideoEnergy  **************************/


// Structure that encodes a binary MRF on a sequence of (independant) images
// T : int, double, float...
template <class T>
struct VideoVolumeEnergy {

  // data structure
  int nt; // number of frames

  // data structure information already given in each Pot[t]
  int nx; // width
  int ny; // height
  int nz; // and width of each image (all images have the same size)

  int neighborhoodsize; // in the actual version of the code, has to be 6 or 26 (greatly recommanded : 26)


  // energy description
  vector< VolumeEnergy<T> > Frames; // the energies defined independently on each frame. For instance Frames[t].Pot(x,y,z) is the cost, for point (x,y,z), to be part of the foreground organ in image t  -  the cost of being part of the background.

  // energy extras
  double constante; // additional constant to the total energy, which is the sum of the energies of each frame


  // constructors 
  VideoVolumeEnergy(){
    nt = 0; 
    nx = 0; ny = 0; nz = 0;
    constante = 0;
    neighborhoodsize = 26;
  }

  void allocate(int nnt){
    Frames.resize(nnt);
    nt = nnt;
  }

  VideoVolumeEnergy(int nnt){
    allocate(nnt);
    nx = 0; ny = 0; nz = 0;
    constante = 0;
    neighborhoodsize = 26;
  }

  VideoVolumeEnergy(int nnt, int nnx, int nny, int nnz, int neigh = 26){
    allocate(nnt,nnx,nny,nnz,neigh);
  }
  
  void allocate(int nnt, int nnx, int nny, int nnz, int neigh = 26){
    allocate(nnt);
    nx = nnx; ny = nny; nz = nnz;
    constante = 0;
    neighborhoodsize = neigh;
    
    for (int t = 0; t<nt; t++){
      Frames[t].neighborhoodsize = neighborhoodsize;
      Frames[t].allocate(nx, ny, nz);
    }
  }
  
  VideoVolumeEnergy(const vector< VolumeEnergy<T> >& FFrames) : Frames(FFrames){ // attention, copie des pointeurs sur les données
    nt = Frames.size();
    nx = 0; ny = 0; nz = 0; neighborhoodsize = 26; if (nt !=0){ nx = Frames[0].nx; ny = Frames[0].ny; nz = Frames[0].nz; neighborhoodsize = Frames[0].neighborhoodsize;}
    constante = 0;
  }


  ~VideoVolumeEnergy(){} // destructor by default 

  void empty_light(){ // doesn't clean much (useful if initialised by copying pointers to data which should not be deleted)
    Frames.empty();
  }

  void empty_full(){ // if really needed
    for (int t = 0; t<nt; t++)
      Frames[t].empty();
    Frames.empty();
  }


  void display_text() const{
    
    cout << "Video volume energy : " << nt << " frames of " << nx << " x " << ny << " x " << nz << " pixels." << endl;
    cout << "Additional constant : " << constante << endl << endl;
  }

  
};

 
 
 
// Evaluates the energy associated to a state (a choice of all labels)
template <class T>
double evaluate_energy(const VolumeEnergy<T>& A, const Volume<bool>& S){
  double v = A.constante;
  int ns = A.neighborhoodsize/2;
  assert(A.neighborhoodsize == 26);
  assert(S.nx == A.nx);
  assert(S.ny == A.ny);
  assert(S.nz == A.nz);
  for (int i = 0; i<A.nx; i++)
    for (int j = 0; j<A.ny; j++)
      for (int k = 0; k<A.nz; k++){

	if (! S(i,j,k))  // individual potentials
	  v += A.Pot(i,j,k); 
	//      else v += 0
	
	
	// Interaction terms


	/*  // This way was far more too complicated to read/write. Skip to new way.
	int i0, j0, k0; // loop over all possible neighbors (i0,j0,k0) from (i,j,k) in a direction indexed from (i,j,k)
	

	int dz = -1; k0 = k - 1; // At previous depth, look at all pixels :
	int dx, dy;
	if (k0 >= 0){ // neighbor has to be in the volume
	  
	  for (dx = -1; dx <= 1; dx++){
	    i0 = i + dx;
	    if ( (i0<0) || (i0 >= A.nx))  // neighbord has to be in the volume.
	      continue;
	    
	    for (dy = -1; dy <= 1; dy++){
	      j0 = j + dy;
	      if ( (j0<0) || (j0 >= A.ny) )
		continue;
	      
	      if (S(i,j,k) != S(i0,j0,k0))  // count interaction term of neighbors have different labels
		v += A.Inter(i,j,k)[ (dx+1) + (dy+1)*3 + (dz+1)*9 ];
	    }
	  }
	}
	
	dz = 0;  k0 = k; // At current depth,
	dy = -1; j0 = j - 1; // look first only for points on the top
	if ( j0 >= 0 ){
	  for (dx = -1; dx <= 1; dx++){
	    i0 = i + dx;
	    if ( (i0<0) || (i0 >= A.nx))  // neighbord has to be in the volume.
	      continue;
	    
	    if (S(i,j,k) != S(i0,j0,k0))  // count interaction term of neighbors have different labels
	      v += A.Inter(i,j,k)[ (dx+1) + (dy+1)*3 + (dz+1)*9 ];
	  }
	}
	

	dy = 0;  j0 = j; // and also for the centered-point on the right
	dx = 1;  i0 = i+1;
	if ( i0 < A.nx){
	  if (S(i,j,k) != S(i0,j0,k0))  // count interaction term of neighbors have different labels
	    v += A.Inter(i,j,k)[ (dx+1) + (dy+1)*3 + (dz+1)*9 ];
	}
*/


	// Easier way to go through all interacting edges
	for (int d = 0; d<ns; d++){ // loop on neighbors. By (forced) luck, the 13 smallest indices within the 26 neighbors form precisely the good hald of the links to consider.
	  int dx = (d%3) - 1 ; int dy = ((d%9)/3) - 1; int dz = (d/9) -1;  // supposing neighborhood size = 26 (Rubik's cube) 
	  
	  // who is my neighbor ?
	  int i0 = i+dx; int j0 = j+dy; int k0 = k+dz;

	  if ( (i0 >= 0) && (j0 >= 0) && (k0 >= 0) && (i0 < A.nx) && (j0 < A.ny) && (k0 < A.nz) ) { // most of these tests are useless
	    if (S(i,j,k) != S(i0,j0,k0))  // count interaction term of neighbors have different labels
	      v += A.Inter(i,j,k)[ d ];
	  }
	}
      }
  return v;
}
 

// Evaluates the energy associated to a state (a choice of all labels)
template <class T>
double evaluate_energy(const VideoVolumeEnergy<T>& A, const vector< Volume<bool> >& S){
  assert(A.nt == (int) S.size() );
  double v = A.constante;
  for (int k = 0; k<A.nt; k++)
    v += evaluate_energy(A.Frames[k], S[k]);
  return v;
}


 




/************************************************************************************/
/***************** Coupled segmentation of several 3D sequences  ********************/
/************************************************************************************/


// Evaluates the energy associated to a state (a choice of all labels)
// first vector< ... >  : denotes sequence type (MR, CT, etc.)
// second vector< ... > : stands for time sequence
template <class T>
double evaluate_energy(const vector< VideoVolumeEnergy<T> >& A, const vector< vector< Volume<bool> > >& S){
  double v = 0;
  for (unsigned int k = 0; k<A.size(); k++)
    v += evaluate_energy(A[k], S[k]);
  return v;
}




} // end of namespace

#endif

